#include "OperatingSystem.h"
#include "Interface.h"
#include "Hero.h"

/* This one is purely procedural. OO is eschewed until whole object system
 * is redone in DD fashion. Then those will be members of shObject.
 */
/* shObjectIlks are all merged but I haven't kept my promise. Ehh. -- MB */

int systemHackingModifier(int sys)
{
    return sys;
}

/* Hardware must be supported by software. */
int systemSupportsInterface (int sys, shCompInterface interface)
{
    switch (interface)
    {
        case kIVisualOnly: /* Taken for granted. */
            return 1;
        case kIVisualVoice:
            return sys > kAbysmalSys;
        case kIUniversal:
            return sys > kPoorSys;
        default:
            return 0;
    }
}

int hasSystem (shObject *computer)
{
    if (computer->mEnhancement != kNoSys)
        return 1;
    else {
        if (!Hero.isBlind ()) {
            I->p ("DISK BOOT FAILURE. INSERT SYSTEM DISK AND PRESS ENTER.");
            computer->setEnhancementKnown ();
            I->pause ();
            I->p ("Your computer lights floppy drive LEDs for a moment.");
        } else
            I->p ("You hear floppy drives seeking briefly.");
        return 0;
    }
}

int systemRunPOST (shObject *computer)
{
    int sys = computer->mEnhancement;
    if (sys >= kGoodSys) {
        if (!computer->isBugginessKnown ()) {
            if (computer->isBuggy ())
                I->p ("Warning! POST detected hardware malfunction.");
            else if (computer->isOptimized ())
                I->p ("POST running... high performance settings active.");
            else
                I->p ("POST running... all tests passed.");
            computer->setBugginessKnown ();
            if (computer->isBuggy ())
                return I->yn ("Cancel execution?");
        }
    }
    return 1;
}

/* Returns 1 if system check passed and 0 otherwise. */
int systemCheckDisk (shObject *computer, shObject *disk)
{
    int sys = computer->mEnhancement;
    if (sys >= kGoodSys && !disk->isIlkKnown () &&
        (disk->isA (kObjSpamDisk) || disk->isA (kObjHypnosisDisk)))
    {
        I->p ("Program has been recognized as fraudulent software.");
        disk->setIlkKnown ();
        I->p ("%c - %s", disk->mLetter, disk->inv ());
        int ans = I->yn ("Run anyway?");
        if (ans == 0) return 0;
    }

    /* Does not tell debugged from optimized. That would be too good. */
    if (sys >= kExcellentSys
        && !disk->isBugginessKnown () && disk->isBuggy ())
    {
        I->p ("Software has not passed reliability tests.");
        disk->setBugginessKnown ();
        I->p ("%c - %s", disk->mLetter, disk->inv ());
        int ans = I->yn ("Continue?");
        if (ans == 0) return 0;
    }

    /* This one does. */
    if (sys >= kSuperbSys && !disk->isBugginessKnown ())
    {
        disk->setBugginessKnown ();
        I->p ("%c - %s", disk->mLetter, disk->inv ());
        int ans = I->yn ("Continue?");
        if (ans == 0) return 0;
    }

    /* Antivirus in computer can recoved infected floppies. */
    if (computer->isFooproof () && disk->isInfected ()) {
        I->p ("Antivirus alert: trojan Zap'M-OverwriteDisk found.");
        disk->setDamageKnown ();
        if (disk->isA (kObjComputerVirus)) {
            disk->setIlkKnown ();
        }
        I->p ("%c - %s", disk->mLetter, disk->inv ());
        if (I->yn ("Recover floppy disk?", disk->mCount > 1 ? "s" : "")) {
            disk = Hero.removeOneObjectFromInventory (disk);
            disk->resetInfected ();
            if (disk->isA (kObjComputerVirus)) {
                disk->mIlkId = kObjBlankDisk;
            }
            Hero.addObjectToInventory (disk, 0);
            return 0; /* Disinfection should take time. */
        } else { /* No point in disallowing that. */
            int ans = I->yn ("Run anyway? (computer will not get infected)");
            if (ans == 0) return 0;
        }
    }

    return 1; /* Nothing suspicious or warnings ignored. */
}

/* In each five minutes system will reserve 12d6 seconds for updates. */
#define INTERVAL 300000
#define UPDLENGTH NDX(12, 6)
// RNG argument = interval / 1000 - max(updlength) - 1
int beginUpdate = RNG(227);
int endUpdate = beginUpdate + UPDLENGTH;
/* If you change any values remember to update the routine below. */
void resetUpdatePeriod(void)
{
    beginUpdate = RNG(227);
    endUpdate = beginUpdate + UPDLENGTH;
}

/* Returns 1 if OS has decided to troll you right now and 0 otherwise. */
int isUpdateTime (int sys)
{ /* Good system does not lock up your computer during updates. */
    if (sys <= kAverageSys
        && ((Clock % INTERVAL) / 1000 >= beginUpdate)
        && ((Clock % INTERVAL) / 1000 < endUpdate)) {
        return 1;
    }
    return 0;
}

void systemUpdateMessage (void)
{
    int maxUpdates = endUpdate - beginUpdate;
    int updateNumber = (Clock % INTERVAL) / 1000 - beginUpdate + 1;
    /* Hint! Hint! XD */
    I->p ("Do not turn off or kick your computer. Loading update %d of %d.",
        updateNumber, maxUpdates);
}
